// Scintilla source code edit control
// @file LexPB.cxx
// Lexer for PowerBasic by Roland Walter, roland@rowalt.de (for PowerBasic see www.powerbasic.com)
// Lexer for PowerBasic updated by Juergen Kuehlwein
//
// Changes:
// 17.10.2003: Toggling of subs/functions now until next sub/function - this gives better results
// 29.10.2003: 1. Bug: Toggling didn't work for subs/functions added in editor
//             2. Own colors for PB constants and Inline Assembler SCE_B_CONSTANT and SCE_B_ASM
//             3. Several smaller syntax coloring improvements and speed optimizations
// 12.07.2004: 1. Toggling for macros added
//             2. Further folding speed optimitations (for people dealing with very large listings)
// 21.03.2014  improved code coloring, enhanced code folding
//
// Necessary changes for the PB lexer in Scintilla project:
//  - In SciLexer.h and Scintilla.iface:
//
//    #define SCLEX_POWERBASIC 51       //ID for PowerBasic lexer
//    (...)
//    #define SCE_B_DEFAULT 0           // PB lexer
//    #define SCE_B_COMMENT 1           // PB lexer
//    #define SCE_B_NUMBER 2            // PB lexer
//    #define SCE_B_KEYWORD 3           // PB lexer
//    #define SCE_B_KEYWORD2 10         // PB lexer
//    #define SCE_B_STRING 4            // PB lexer
//    #define SCE_B_PREPROCESSOR 5      // PB lexer
//    #define SCE_B_OPERATOR 6          // PB lexer
//    #define SCE_B_IDENTIFIER 7        // PB lexer
//    #define SCE_B_CONSTANT 13         // PB lexer
//    #define SCE_B_ASM 14              // PB lexer        

//  - Statement added to KeyWords.cxx:      'LINK_LEXER(lmPB);'
//  - Statement added to scintilla_vc6.mak: '$(DIR_O)\LexPB.obj: ...\src\LexPB.cxx $(LEX_HEADERS)'
//
// Copyright for Scintilla: 1998-2001 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <ctype.h>

#include "ILexer.h"
#include "Scintilla.h"
#include "SciLexer.h"

#include "WordList.h"
#include "LexAccessor.h"
#include "Accessor.h"
#include "StyleContext.h"
#include "CharacterSet.h"
#include "LexerModule.h"

#include "windows.h"


#ifdef SCI_NAMESPACE
using namespace Scintilla;
#endif


//**********************************************************************************************
// helper functions
//**********************************************************************************************

static inline bool IsTypeCharacter(const int ch) // valid type specifier ?
{
    return ch == '%' || ch == '&' || ch == '@' || ch == '!' || ch == '#' || ch == '$' || ch == '?';
}

static inline bool IsAWordChar(const int ch)     // valid word char ?
{
    return (ch < 0x80) && (isalnum(ch) || ch == '.' || ch == '_');
}

static inline bool IsAWordStart(const int ch)
{
    return (ch < 0x80) && (isalnum(ch) || ch == '_');
}


static inline bool MatchUpperCase(Accessor &styler, int pos, const char *s) {    // upperrcase comparison (a-z,A-Z and space only)
//**********************************************************************************************
// check if *s is at current position (ignore multiple separating spaces)
//**********************************************************************************************
    char prev = 'A';
    char ch;
    for (int i=0; *s; i++) {
        ch = styler.SafeGetCharAt(pos+i);        // get next char from line
        if (ch > 0x60) ch -= '\x20';             // make upper case, if necessary
        if (*s != ch) {                          // if not equal
          if ((ch == ' ') && (prev == ' ')) {    // if prev is a space and current char is a space as well
            goto m1;}                            // iterate anyway
          else {                                 // otherwise
            return false;}}                      // return no match

        s++;                                     // next char to compare
        prev = ch;                               // save current char (might be space)
m1:;
    }
    return true;
}


static inline bool MatchUpperCaseNoSpace(Accessor &styler, int pos, const char *s) {    // upperrcase comparison (a-z,A-Z, no space)
//**********************************************************************************************
// check if *s is at current position (ignore spaces at all)
//**********************************************************************************************
    char ch;
    for (int i=0; *s; i++) {
        ch = styler.SafeGetCharAt(pos+i);        // get next char from line
        if (ch > 0x60) ch -= '\x20';             // make upper case, if necessary
        if (ch == ' ') {goto m2;}                // skip spaces
        if (*s != ch) {return false;}            // if not equal -> return "no match"

        s++;                                     // next char to compare
m2:;
    }
    return true;
}

//**********************************************************************************************
//**********************************************************************************************
// see StyleContext.h and .cxx

// 	void Forward() {                      -> goto next segement (word ?)
// 	void Forward(int nb) {                -> procede nb segements
//	void ForwardBytes(int nb) {           -> procede nb chars
//	void ForwardSetState(int state_) {    -> goto next line and set next line to this state
//	int GetRelative(int n) {              -> get char relative to current pos (may be negative -> get preceding)

// maybe own function -> read to eol

//_itoa(pflag, m, 10);
//OutputDebugStringA(m);
//	  OutputDebugStringA("prefix \"asm\"");}

//**********************************************************************************************
//**********************************************************************************************





static void ColourisePBDoc(unsigned int startPos, int length, int initStyle, WordList *keywordlists[], Accessor &styler) {
//**********************************************************************************************
// scan text char by char, set style (SCE_B_...), a style is valid until it is changed 
//**********************************************************************************************
    WordList &keywords  = *keywordlists[0];         // Power Basic Keywords
    WordList &keywords2 = *keywordlists[1];         // Power Basic meatstatements

	int prev   = 0;                                   // preceding char
	bool pflag = false;                               // prefix -> = 1, "asm " -> = 2, reset to 0 for "end prefix" 


    styler.StartAt(startPos);
    StyleContext sc(startPos, length, initStyle, styler);


    for (; sc.More(); sc.Forward()) {            // scan chars
      switch (sc.state) {                       // switch states (handle all except identifier and default state)
        case SCE_B_OPERATOR: {                // operators are all of length 1
          sc.SetState(SCE_B_DEFAULT);      // reset to default afterwards
          break; }


        case SCE_B_KEYWORD: {
          if (!IsAWordChar(sc.ch)) {        // if not a word char (might be a space)
            if (!IsTypeCharacter(sc.ch)) {//if not a type specifier
              char s[100];             // asciiz * 100
              sc.GetCurrentLowered(s, sizeof(s));          // get segment (lcase word) before current position

              if (keywords.InList(s)) { //if in keyword list
                if (strcmp(s, "rem") == 0) {              // if "rem"
                   sc.ChangeState(SCE_B_COMMENT);
                   if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT); }}

                else if (strcmp(s, "asm") == 0) {          // if "asm"
                   sc.ChangeState(SCE_B_ASM);
                   if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT); }} 

                else if (strcmp(s, "prefix") == 0) {      // if "prefix"
                   if (MatchUpperCaseNoSpace(styler, sc.currentPos+1 ,"\"ASM\"")) 
                     {pflag = true; sc.SetState(SCE_B_PREFIXASM); sc.ForwardSetState(SCE_B_PREFIXASM);}
                   else if (MatchUpperCaseNoSpace(styler, sc.currentPos+1 ,"\"!\""))  
                     {sc.SetState(SCE_B_PREFIXASM); sc.ForwardSetState(SCE_B_PREFIXASM);}
                   else {sc.SetState(SCE_B_DEFAULT); }} 

                else if (strcmp(s, "#utility") == 0) {    //if "#utility" (preprocessor statement)
                  sc.ChangeState(SCE_B_PREPROCESSOR);
                  if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT); }}

                else {
                  sc.SetState(SCE_B_DEFAULT); }}         // it was a keyword, reset state afterwards to default

              else if (keywords2.InList(s)) {              // it wasnt a keyword, so check for keyword2
                sc.ChangeState(SCE_B_KEYWORD2); }

              else {                    // it wasnt a keyword or keyword2, so it must have been an identifier (nornal text)
                sc.ChangeState(SCE_B_IDENTIFIER);        // change to identifier
                sc.SetState(SCE_B_DEFAULT); }}}             // following are default state again

          break; }


        case SCE_B_NUMBER: {
          if (!IsAWordChar(sc.ch)) {sc.SetState(SCE_B_DEFAULT);} // reset to default if not a word char
          break; }


        case SCE_B_STRING: {                  // we are inside a literal string
          if (sc.ch == '\"'){sc.ForwardSetState(SCE_B_DEFAULT);}
          if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT);}
          break; }


        case SCE_B_CONSTANT: {
          if (!IsAWordChar(sc.ch)) {sc.SetState(SCE_B_DEFAULT);} // reset to default if not a word char
          break; }


        case SCE_B_COMMENT: {
          if (sc.atLineEnd) {if (pflag) {sc.SetState(SCE_B_PREFIXASM); sc.ForwardSetState(SCE_B_PREFIXASM);} else {sc.SetState(SCE_B_DEFAULT);}}
          break; }


        case SCE_B_KEYWORD2: {
          if (sc.ch == '\"'){sc.SetState(SCE_B_STRING);}       // might be followed by a string
          if (sc.ch == '\''){sc.SetState(SCE_B_COMMENT);}      // or a comment

          if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT);}      // reset to default at line end
          break; }


        case SCE_B_PREPROCESSOR: {
          if (sc.ch == '\''){sc.SetState(SCE_B_COMMENT);}
          if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT);}      // reset to default at line end
          break; }


        case SCE_B_ASM: {
          if (sc.ch == ';'){sc.SetState(SCE_B_COMMENT);}         // asm comment sp_show_
          if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT);}
          break; }


        case SCE_B_PREFIXASM: {
          pflag = true;
          if (sc.ch != ' ') {             // if not a space (any following char should be "end ", otherwise still ASM)

            if (((sc.GetRelative(0) & 95) == 'E') && ((sc.GetRelative(1) & 95) == 'N')) {
              if (MatchUpperCase(styler, sc.currentPos, "END PREFIX")) { 
				  pflag = false; sc.SetState(SCE_B_DEFAULT);}}}

		    if (sc.ch == ';') {sc.SetState(SCE_B_COMMENT);}         // asm comment 
            if (sc.atLineEnd) {sc.SetState(SCE_B_PREFIXASM);}
            break; }

        }                                        // end switch (sc.state)



//**********************************************************************************************
// if not in string, check for "_" and preceding char (switch following to comment mode) 
//**********************************************************************************************
        if (sc.state != SCE_B_STRING) {
          if (sc.ch == '_') {                    // if line continuation char found and not inside a literal string
                                                 // -> check preceding char
            if (prev < 48) {sc.ForwardSetState(SCE_B_COMMENT);}      // everything below "0" is ok

            if (prev == ':') {sc.ForwardSetState(SCE_B_COMMENT);}    // and all these are ok
            if (prev == ';') {sc.ForwardSetState(SCE_B_COMMENT);} 
            if (prev == '<') {sc.ForwardSetState(SCE_B_COMMENT);} 
            if (prev == '=') {sc.ForwardSetState(SCE_B_COMMENT);} 
            if (prev == '>') {sc.ForwardSetState(SCE_B_COMMENT);} 
            
            if (prev == '[') {sc.ForwardSetState(SCE_B_COMMENT);} 
            if (prev == 92)  {sc.ForwardSetState(SCE_B_COMMENT);}    // backslash
            if (prev == ']') {sc.ForwardSetState(SCE_B_COMMENT);} 
            if (prev == '^') {sc.ForwardSetState(SCE_B_COMMENT);} 
          }
        }


//**********************************************************************************************
// we are in default state, determine if a new state should be entered:
//**********************************************************************************************
        if (sc.state == SCE_B_DEFAULT) {
          if (sc.ch == '\'') {sc.SetState(SCE_B_COMMENT);}
          else if (sc.ch == '\"') {sc.SetState(SCE_B_STRING);}
          else if (sc.ch == '&' && tolower(sc.chNext) == 'h') {sc.SetState(SCE_B_NUMBER);}
          else if (sc.ch == '&' && tolower(sc.chNext) == 'b') {sc.SetState(SCE_B_NUMBER);}
          else if (sc.ch == '&' && tolower(sc.chNext) == 'o') {sc.SetState(SCE_B_NUMBER);}
          else if (IsADigit(sc.ch) || (sc.ch == '.' && IsADigit(sc.chNext))) {sc.SetState(SCE_B_NUMBER);}
          else if (IsAWordStart(sc.ch)) {sc.SetState(SCE_B_KEYWORD);} // make every new word a keyword first
          else if (sc.ch == '%') {sc.SetState(SCE_B_CONSTANT);}
          else if (sc.ch == '$') {sc.SetState(SCE_B_CONSTANT);}
          else if (sc.ch == '#') {sc.SetState(SCE_B_KEYWORD);}
          else if (sc.ch == '!') {sc.SetState(SCE_B_ASM);}
          else if (isoperator(static_cast<char>(sc.ch)) || (sc.ch == '\\')) {sc.SetState(SCE_B_OPERATOR);}
        }
        prev = sc.ch;                            // save current char for next iteration

    }                                            // next
    sc.Complete();
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// rules:

// if level <= SC_FOLDLEVELBASE then level for these is SC_FOLDLEVEL - 1
// if level > SC_FOLDLEVELBASE then level for these is level + 1
// this way you can enclose SC_FOLDLEVELBASE lines (sub/function .., type enum
//   '->    '+*....   increase fold level (set a userdefined folding point, must be left aligned, no leading spaces)
//   '->    '-*....   decrease fold level

//   sub/(thread/callback)function/fastproc/class/type/union/enum/asmdata/macro(multiline) and corresponding end 
//   are always at SC_FOLDLEVELBASE level (en)d (sub, function, class, type, union, enum, asmdata, macro)
//   'exclude "function =", "method =", property =

//         '#if
//         '#else(if)
//         '#endif         // dont fold at those

//   '#pbforms (b)egin ...
//   '#pbforms (e)nd ...

//   'interface, (class/override) method, property (set/get)

//   'try

//   'prefix

//   'select

//   'if

//   'for, do, while (1. word)
//   'next, loop, wend

//   'end -> level = level - 1 (except for end (se)lect)


//   'be careful -> watch out for and handle single line: property, method, macro, if, for, while, do  
//    -> set flag for these:  0 = normal, 1 = if, 2 = for etc.     reset if new line and set level


//    scan remainder of line, look for literal strings, comments (continuation char as well) and line end

//    property/method -> look for end property/end method
//    macro           -> look for "="
//    for/while/do    -> look for next/wend/loop
//    if              -> look for "then", and look for other statements 


//**********************************************************************************************
// fold level is a number between 0 and 0xFFF ("& SC_FOLDLEVELNUMBERMASK" isolates this level from style byte)
// "| SC_FOLDLEVELHEADERFLAG" adds a fold point to this line
// each line needs to be set a folding level, not only fold points (or headerflag)
//	SC_FOLDLEVELWHITEFLAG 0x1000    line contains nothing (but spaces or tabs)
//**********************************************************************************************


static void FoldPBDoc(unsigned int startPos, int length, int, WordList *[], Accessor &styler) {
//**********************************************************************************************
// handle folding
//**********************************************************************************************
if( styler.GetPropertyInt("fold") == 0 ) return; // No folding enabled, no reason to continue...


char chU;                                        // current char (uppercase)
char chnU;                                       // next char (uppercase)
unsigned int header = 0;                         // set header flag ?
unsigned int first  = startPos;                  // is it the first pos in line ?

unsigned int endPos = startPos + length;         // scan until here
int lineCurrent     = styler.GetLine(startPos);  // current line # 
int lc              = SC_FOLDLEVELBASE;          // init to default


    if (lineCurrent > 0) {
		lc = styler.LevelAt(lineCurrent) & SC_FOLDLEVELNUMBERMASK; }       //get current level, mask out flags
    int ln = lc;


char ch = ' ';
char chNext   = styler[startPos];                // first char
char prev = 'A';
bool fNewLine = true;                            // flags
bool qflag    = false;                           // within quotes flag
int  flag     = 0;                               // flag for special handling (if=1, for=2, while=3, do=4, macro=5
                                                 // property=6, method=7, 0=no extra handling)

//char m[100] = "initial:                     " ;
//char m1[100];
//_itoa(lc, &m[10] , 10);                          //current level
//_itoa(lineCurrent, m1 , 10);                     //current line
//strcat(m, "  -  ");
//strcat(m, m1);
//OutputDebugStringA(m);                           //test


//********************************************************************************
// main for/next loop
//********************************************************************************


    for (unsigned int i = startPos; i < endPos; i++) {               // loop through chars
      ch = chNext;                          // current char
      chNext = styler.SafeGetCharAt(i + 1);      // next char
      chU  = ch & 95;                            // current char (uppercase)
      chnU = chNext & 95;                        // next char (uppercase)


      if (fNewLine) {                            // Begin of a new line, the first non blank is the first token
//OutputDebugStringA("newline");
        if (ch == ' ')  {goto l1; }              // skip spaces
        if (ch == '\t') {goto l1; }              // skip tabs

        fNewLine = false;                        // now we are at the first token

        if (ch == '\'')  {
          if (i == first) {                      // only, if first in line (only leftaligned, ignore second and further pos)
            if (styler.SafeGetCharAt(i + 2) == '*') {
              if (chNext == '+') {               // "'+*..." (user fold start)
                if (lc > SC_FOLDLEVELBASE) {     //(regular) user fold point
                  ln = lc + 1; }
                else {                           //at SC_FOLDLEVELBASE, step back onr level (enclose SC_FOLDLEVELBASE points)
                  lc = SC_FOLDLEVELBASE - 1;
                  ln = lc + 1; }

				header = SC_FOLDLEVELHEADERFLAG;
			    goto l1; }

              else if (chNext == '-') {          // "'-*..." (user fold end)
                ln = lc - 1;
			    goto l1; }}}}


        else if (ch == '#') {                    // #PBFORMS ?
          if (chnU == 'P') {
            if (MatchUpperCaseNoSpace(styler, i+2, "BFORMSBEGIN")) {
              header = SC_FOLDLEVELHEADERFLAG;
              ln = lc + 1;}
		    else if (MatchUpperCaseNoSpace(styler, i+2, "BFORMSEND")) {
              ln = lc - 1; }}}


//**********************************************************************************************
// check first nonspace char (A-Z uppercase)
//**********************************************************************************************

		
// always at SC_FOLDLEVELBASE level   sub/function etc. -> ok
// 'interface, (class/override) method, property (set/get) -> ok
// try, prefix -> ok

// select -> ok
// if, for, do, while (1. word) -> ok
// next, loop, wend -> ok

//**********************************************************************************************

        switch (chU) {                      
          case 'E': {
            if (chnU == 'N') {
              if (MatchUpperCase(styler,i,"END ")) {
                ln = lc - 1; 
                goto l1; } 
              if (MatchUpperCase(styler,i,"ENUM ")) {
                header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                lc = SC_FOLDLEVELBASE;         //enum always at base level
                ln = lc + 1;
                goto l1;}}
            break; }

			
          case 'F': {
            if (chnU == 'U') {
              if (MatchUpperCase(styler,i,"FUNCTION ")) {
                if (MatchUpperCaseNoSpace(styler,i,"FUNCTION=")) {}  //must not be "Function = ..."
                else {    
					
//					OutputDebugStringA("function");
                  header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                  lc = SC_FOLDLEVELBASE;         //functions are all at base level
                  ln = lc + 1;
                  goto l1; }}}

            if (chnU == 'A') {
              if (MatchUpperCase(styler,i,"FASTPROC ")) {
                if (MatchUpperCaseNoSpace(styler,i,"FASTPROC=")) {}  //must not be "Fastproc = ..."
                else {    
					
//					OutputDebugStringA("function");
                  header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                  lc = SC_FOLDLEVELBASE;         //functions are all at base level
                  ln = lc + 1;
                  goto l1; }}}

            if (chnU == 'O') {
              if (MatchUpperCase(styler,i,"FOR ")) {
                header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                ln = lc + 1;
                flag = 2;
				goto l1; }}

            break; }


          case 'I': {             
            if (chnU == 'F') {
              if (styler.SafeGetCharAt(i+2) == ' ') {

//int t = ch;
////m = m + chU;
//_itoa(lc, m, 10);
////OutputDebugStringA(&ch);    //geht                   drawcircle  foldpoint
////		m[0]=chU;
//OutputDebugStringA(m);    //geht




                header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                ln = lc + 1; 
                flag = 1; 
                goto l1; }}

            if (chnU == 'N') {
              if (MatchUpperCase(styler,i,"INTERFACE ")) {
                header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                ln = lc + 1; 
				goto l1; }}

            break; }



          case 'S': {
            if (chnU == 'U') {
              if (MatchUpperCase(styler,i,"SUB ")) {
                header = SC_FOLDLEVELHEADERFLAG; // fold point
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                goto l1;}}

            if (chnU == 'E') {
              if (MatchUpperCase(styler,i,"SELECT ")) {
                header = SC_FOLDLEVELHEADERFLAG; // fold point
                ln = lc + 1;
                goto l1;}}

            break; }


          case 'C': {
            if (chnU == 'A') {
              if (MatchUpperCase(styler,i,"CALLBACK ")) {            //fold point
                header = SC_FOLDLEVELHEADERFLAG;
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                goto l1; }}

            if (chnU == 'L') {
               if (MatchUpperCase(styler,i,"CLASS ")) {
                 if (MatchUpperCaseNoSpace(styler,i,"CLASSMETHOD")) {
                   if (MatchUpperCaseNoSpace(styler,i,"METHOD=")) {} //must not be "Method = ..."
                   else {    
                     header = SC_FOLDLEVELHEADERFLAG; //fold point
                     ln = lc + 1;
                     flag = 7;
					 goto l1; }}
                 else {
                   header = SC_FOLDLEVELHEADERFLAG;                    //fold point
                   lc = SC_FOLDLEVELBASE;          //functions are all at base level
                   ln = lc + 1;
				   goto l1; }}}

            break; }


          case 'T': {
            if (chnU == 'H') {
              if (MatchUpperCaseNoSpace(styler,i,"THREADFUNCTION")) {
                header = SC_FOLDLEVELHEADERFLAG; //fold point
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                goto l1; }}

            if (chnU == 'Y') {
               if (MatchUpperCase(styler,i,"TYPE ")) {


//_itoa(lc, m, 10);
////OutputDebugStringA(&ch);   
////		m[0]=chU;
//OutputDebugStringA(m);   

                header = SC_FOLDLEVELHEADERFLAG; //fold point
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                goto l1; }}
           
            if (chnU == 'R') {
               if (MatchUpperCase(styler,i,"TRY")) {
                 if ((styler.SafeGetCharAt(i+3) == ' ') || (styler.SafeGetCharAt(i+3) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG; //fold point
                   ln = lc + 1;
				   goto l1; }}}

            break; }


          case 'U': {
            if (chnU == 'N') {
               if (MatchUpperCase(styler,i,"UNION ")) {
                header = SC_FOLDLEVELHEADERFLAG; //fold point
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                goto l1; }}
           
            break; }


          case 'W': {
            if (chnU == 'H') {
               if (MatchUpperCase(styler,i,"WHILE ")) {
                header = SC_FOLDLEVELHEADERFLAG; //fold point
                ln = lc + 1;
                flag = 4;
                goto l1; }}
           
            if (chnU == 'E') {
               if (MatchUpperCase(styler,i,"WEND")) {
                 if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                   ln = lc - 1;
                   goto l1; }}}

            break; }


          case 'L': {
            if (chnU == 'O') {
               if (MatchUpperCase(styler,i,"LOOP")) {
                 if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                   ln = lc - 1;
				   goto l1; }}}

            break; }


          case 'N': {
            if (chnU == 'E') {
               if (MatchUpperCase(styler,i,"NEXT")) {
                 if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                   ln = lc - 1;
				   goto l1; }}}

            break; }


          case 'D': {
            if (chnU == 'O') {
              if ((styler.SafeGetCharAt(i+2) == ' ') || (styler.SafeGetCharAt(i+2) == '\r')) {
                header = SC_FOLDLEVELHEADERFLAG; //fold point
                ln = lc + 1;
                flag = 3;
                goto l1; }}
           
            break; }


          case 'M': {
            if (chnU == 'A') {
              if (MatchUpperCase(styler,i,"MACRO ")) {
                header = SC_FOLDLEVELHEADERFLAG; // fold point
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                flag = 5;
                goto l1;}}

            if (chnU == 'E') {
              if (MatchUpperCase(styler,i,"METHOD ")) {
                if (MatchUpperCaseNoSpace(styler,i,"METHOD=")) {} //must not be "Method = ..."
                else {    
                  header = SC_FOLDLEVELHEADERFLAG; // fold point
                  ln = lc + 1;
                  flag = 7;
				  goto l1;}}}

            break; }


          case 'O': {
            if (chnU == 'V') {
              if (MatchUpperCaseNoSpace(styler,i,"OVERRIDEMETHOD")) {
                if (MatchUpperCaseNoSpace(styler,i,"METHOD=")) {} //must not be "Method = ..."
                else {    
                  header = SC_FOLDLEVELHEADERFLAG; //fold point
                  ln = lc + 1;
                  flag = 7;
				  goto l1; }}

              else if (MatchUpperCaseNoSpace(styler,i,"OVERRIDEPROPERTY")) {
                if (MatchUpperCaseNoSpace(styler,i,"PROPERTY=")) {}  //must not be "Property = ..."
                else {    
                  header = SC_FOLDLEVELHEADERFLAG;                    //fold point
                  ln = lc + 1;
                  flag = 6;
				  goto l1; }}}
           
            break; }


          case 'P': {
            if (chnU == 'R') {
              if (MatchUpperCase(styler,i,"PROPERTY ")) {
                if (MatchUpperCaseNoSpace(styler,i,"PROPERTY=")) {}  //must not be "Property = ..."
                else {    
                  header = SC_FOLDLEVELHEADERFLAG; // fold point
                  ln = lc + 1;
                  flag = 6;
                  goto l1;}}

              if (MatchUpperCase(styler,i,"PREFIX ")) {
                header = SC_FOLDLEVELHEADERFLAG; // fold point
                ln = lc + 1;
                goto l1;}}

            break; }



          case 'A': {
            if (chnU == 'S') {
              if (MatchUpperCase(styler,i,"ASMDATA ")) {
                header = SC_FOLDLEVELHEADERFLAG; // fold point
                lc = SC_FOLDLEVELBASE;           //functions are all at base level
                ln = lc + 1;
                goto l1;}}
            break; }


          }                                        // end switch (ch)
        }                                          // end if( fNewLine )


//**********************************************************************************************
// do more checks based on flag
//**********************************************************************************************
        if (flag == 0) {goto l2;}                //flag = 0 -> just skip this section of code
        else {                                   //check for quotes, if within -> skip following
          if (ch == '\"') {                      //check for literal 
            if (qflag == false) {qflag = true; }
			else {qflag = false; }
			goto l1; }

		  if (qflag == false) {
			if (ch == '\'') {flag = 0; goto l1; }    //comment found -> abort further scanning
		
//**********************************************************************************************
// if not in string, check for "_" and preceding char (following is comment -> abort further scanning) 
//**********************************************************************************************
              if (ch == '_') {                        //line continuation char found (not inside a literal string)
                                                      // -> check preceding char
                if (prev < 48) {ln = lc + 1; flag = 0; goto l1;}      // everything below "0" is ok

                if (prev == ':') {ln = lc + 1; flag = 0; goto l1;}    // and all these are ok
                if (prev == ';') {ln = lc + 1; flag = 0; goto l1;} 
                if (prev == '<') {ln = lc + 1; flag = 0; goto l1;} 
                if (prev == '=') {ln = lc + 1; flag = 0; goto l1;} 
                if (prev == '>') {ln = lc + 1; flag = 0; goto l1;} 
            
                if (prev == '[') {ln = lc + 1; flag = 0; goto l1;} 
                if (prev == 92)  {ln = lc + 1; flag = 0; goto l1;}    // backslash
                if (prev == ']') {ln = lc + 1; flag = 0; goto l1;} 
				if (prev == '^') {ln = lc + 1; flag = 0; goto l1;}}


//**********************************************************************************************
// handles special cases
//**********************************************************************************************
			switch (flag) {                          //handle flagged lines
			  case 1 : {                             //if
				if (chU == 'N') {
                  if (MatchUpperCase(styler,i-3,"THEN")) {     // "N" of "THEN" reached
                    if ((styler.SafeGetCharAt(i+1) == ' ') || (styler.SafeGetCharAt(i+1) == '\r')) {  //space or $cr afterwards
                      if ((styler.SafeGetCharAt(i-4) == ' ') || (styler.SafeGetCharAt(i+4) == ')')) { //space or bracket before
                        flag = 8;
                        goto l1; }}}}

                break; }


			  case 2 : {                             //for
				if ((chU == 'N') &&	(chnU == 'E')) {
                  if (MatchUpperCase(styler,i,"NEXT")) {
                    if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                      ln = lc;
                      flag = 0;
                      goto l1; }}}

                break; }


			  case 3 : {                             //do
				if ((chU == 'L') &&	(chnU == 'O')) {
                  if (MatchUpperCase(styler,i,"LOOP")) {
                    if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                      ln = lc;
                      flag = 0;
                      goto l1; }}}

                break; }



			  case 4 : {                             //while
				if ((chU == 'W') &&	(chnU == 'E')) {
                  if (MatchUpperCase(styler,i,"WEND")) {
                    if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                      ln = lc;
                      flag = 0;
                      goto l1; }}}

                break; }
					   


			  case 5 : {                             //macro
                if (ch == '=') {ln = lc; flag = 0; goto l1; }
                break; }


			  case 6 : {                             //property
				if ((chU == 'E') &&	(chnU == 'N')) {
                  if (MatchUpperCaseNoSpace(styler,i,"ENDPROPERTY")) {
                    ln = lc;
                    flag = 0;
                    goto l1; }}

                break; }
					   

			  case 7 : {                             //method
				if ((chU == 'E') &&	(chnU == 'N')) {
                  if (MatchUpperCaseNoSpace(styler,i,"ENDMETHOD")) {
                    ln = lc;
                    flag = 0;
                    goto l1; }}

                break; }

					   
			  case 8 : {                             //if - following then
                  if ((ch == ' ') || (ch == '\t')) {goto l1;}
                  if ((ch == '\r') || (ch == '\n')) {goto l2;}
                  else {                             // a non space char found -> single line if
                    ln = lc;
                    flag = 0; }

                break; }


			}}}                                      //end switch (flag)

//**********************************************************************************************
// check for line end
//**********************************************************************************************


        //if (ch == '\r') {                        //carriage return -> line end reached
        //  if (chNext == '\n') {i = i + 1; }      //skip line feed
        //    goto l2; }                           //handle line end


l2: ;
        if (ch == '\n') {                        //line feed -> line end reached
          styler.SetLevel(lineCurrent, ((lc & SC_FOLDLEVELNUMBERMASK) | header));                //set level for current line


//_itoa(lineCurrent, m , 10);
//strcat(m, " - set");
//OutputDebugStringA(m);



          lineCurrent++;                         //increase line #
          lc = ln;                               //current level = next level
          fNewLine = true;                       //a new line starts
          first = i + 1;                         //first = pos of first char in this new line
          flag = 0;                              //reset flag
          qflag = false;                         //reset within quotes flag
          header = 0; }


l1: ;
		prev = ch;
    }                                               // next


 
//int t = ch;    
//_itoa(lineCurrent, m, 10 );
//strcat(m, " - EOL");
//    OutputDebugStringA(m);     // should  deliver a line feed !!!!, if not the current line hasnt been styled
//                               // and needs to be styled still  

    
    
    if (ch != '\n') {                            //last char might not be a line feed, (no level set for this line)
      styler.SetLevel(lineCurrent, ((lc & SC_FOLDLEVELNUMBERMASK) | header)); } //set level for current line


    lineCurrent++;                               //sytle next line (level could be different)
    styler.SetLevel(lineCurrent, (ln & SC_FOLDLEVELNUMBERMASK));     //set level for next line

}


//**********************************************************************************************


static const char * const pbWordListDesc[] = {
    "Keywords",
    0
};

LexerModule lmPB(SCLEX_POWERBASIC, ColourisePBDoc, "powerbasic", FoldPBDoc, pbWordListDesc);
