// Scintilla source code edit control
// @file LexRC.cxx
// Lexer for PowerBasic by Roland Walter, roland@rowalt.de (for PowerBasic see www.powerbasic.com)
// Lexer for PowerBasic updated by Juergen Kuehlwein
//
// Changes:
// 26.03.2014  first Version by Juergen Kuehlwein

// Necessary changes for the RC lexer in Scintilla project:
//  - In SciLexer.h and Scintilla.iface:
//
//    #define SCLEX_PC        120       //ID for RC lexer
//    (...)
//    #define SCE_B_DEFAULT 0           //in both VB and PB lexer
//    #define SCE_B_COMMENT 1           //in both VB and PB lexer
//    #define SCE_B_NUMBER 2            //in both VB and PB lexer
//    #define SCE_B_KEYWORD 3           //in both VB and PB lexer
//    #define SCE_B_KEYWORD2 10         //in both VB and PB lexer
//    #define SCE_B_STRING 4            //in both VB and PB lexer
//    #define SCE_B_OPERATOR 6          //in both VB and PB lexer
//    #define SCE_B_IDENTIFIER 7        //in both VB and PB lexer



//    #define SCE_B_PREPROCESSOR 5      //VB lexer only, not in PB lexer
//    #define SCE_B_DATE 8              //VB lexer only, not in PB lexer
//    #define SCE_B_CONSTANT 13         //PB lexer only, not in VB lexer
//    #define SCE_B_ASM 14              //PB lexer only, not in VB lexer

//  - Statement added to KeyWords.cxx:      'LINK_LEXER(lmRC);'
//  - Statement added to scintilla_vc6.mak: '$(DIR_O)\LexRC.obj: ...\src\LexRC.cxx $(LEX_HEADERS)'
//  - Statement added to Catalogue.cxx      'LINK_LEXER(lmRC);
 
//
// Copyright for Scintilla: 1998-2001 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <ctype.h>

#include "ILexer.h"
#include "Scintilla.h"
#include "SciLexer.h"

#include "WordList.h"
#include "LexAccessor.h"
#include "Accessor.h"
#include "StyleContext.h"
#include "CharacterSet.h"
#include "LexerModule.h"

#include "windows.h"


#ifdef SCI_NAMESPACE
using namespace Scintilla;
#endif


//**********************************************************************************************
// helper functions
//**********************************************************************************************

static inline bool IsAWordChar(const int ch)     // valid word char ?
{
    return (ch < 0x80) && (isalnum(ch) || ch == '.' || ch == '_');
}

static inline bool IsAWordStart(const int ch)
{
    return (ch < 0x80) && (isalnum(ch) || ch == '_');
}


static inline bool MatchUpperCase(Accessor &styler, int pos, const char *s) {    // upperrcase comparison (a-z,A-Z and space only)
//**********************************************************************************************
// check if *s is at current position (ignore multiple separating spaces)
//**********************************************************************************************
    char prev = 'A';
    char ch;
    for (int i=0; *s; i++) {
        ch = styler.SafeGetCharAt(pos+i);        // get next char from line
        if (ch > 0x60) ch -= '\x20';             // make upper case, if necessary
        if (*s != ch) {                          // if not equal
          if ((ch == ' ') && (prev == ' ')) {    // if prev is a space and current char is a space as well
            goto m1;}                            // iterate anyway
          else {                                 // otherwise
            return false;}}                      // return no match

        s++;                                     // next char to compare
        prev = ch;                               // save current char (might be space)
m1:;
    }
    return true;
}


static inline bool MatchUpperCaseNoSpace(Accessor &styler, int pos, const char *s) {    // upperrcase comparison (a-z,A-Z, no space)
//**********************************************************************************************
// check if *s is at current position (ignore spaces at all)
//**********************************************************************************************
    char ch;
    for (int i=0; *s; i++) {
        ch = styler.SafeGetCharAt(pos+i);        // get next char from line
        if (ch > 0x60) ch -= '\x20';             // make upper case, if necessary
        if (ch == ' ') {goto m2;}                // skip spaces
        if (*s != ch) {return false;}            // if not equal -> return "no match"

        s++;                                     // next char to compare
m2:;
    }
    return true;
}


//**********************************************************************************************
//**********************************************************************************************
// see StyleContext.h and .cxx

// 	void Forward() {                      -> goto next segement (word ?)
// 	void Forward(int nb) {                -> procede nb segements
//	void ForwardBytes(int nb) {           -> procede nb chars
//	void ForwardSetState(int state_) {    -> goto next line and set next line to this state
//	int GetRelative(int n) {              -> get char relative to current pos (may be negative -> get preceding)

// maybe own function -> read to eol
//**********************************************************************************************
//**********************************************************************************************



static void ColouriseRCDoc(unsigned int startPos, int length, int initStyle, WordList *keywordlists[], Accessor &styler) {
//**********************************************************************************************
// scan text char by char, set style (SCE_B_...), a style is valid until it is changed 
//**********************************************************************************************
    WordList &keywords  = *keywordlists[0];         // RC Keywords
    WordList &keywords2 = *keywordlists[1];         // RC Preprocessor (meatstatements)

    styler.StartAt(startPos);
    StyleContext sc(startPos, length, initStyle, styler);


    for (; sc.More(); sc.Forward()) {            // scan chars
      switch (sc.state) {                       // switch states (handle all except identifier and default state)
        case SCE_B_OPERATOR: {                // operators are all of length 1
          sc.SetState(SCE_B_DEFAULT);      // reset to default afterwards
          break; }


        case SCE_B_KEYWORD: {
          if (!IsAWordChar(sc.ch)) {        // if not a word char (might be a space)
              char s[100];             // asciiz * 100
              sc.GetCurrentLowered(s, sizeof(s));          // get segment (lcase word) before current position

              if (keywords.InList(s)) { //if in keyword list
                sc.SetState(SCE_B_DEFAULT); }         // it was a keyword, reset state afterwards to default

              else if (keywords2.InList(s)) {              // it wasnt a keyword, so check for keyword2
                sc.ChangeState(SCE_B_KEYWORD2);
				sc.SetState(SCE_B_DEFAULT); }         // it was a keyword, reset state afterwards to default
			  
              else {                    // it wasnt a keyword or keyword2, so it must have been an identifier (nornal text)
                sc.ChangeState(SCE_B_IDENTIFIER);        // change to identifier
				sc.SetState(SCE_B_DEFAULT); }}             // following are default state again

          break; }


        case SCE_B_NUMBER: {
          if (!IsAWordChar(sc.ch)) {sc.SetState(SCE_B_DEFAULT);} // reset to default if not a word char
          break; }


        case SCE_B_STRING: {                  // we are inside a literal string
          if (sc.ch == '\"'){sc.ForwardSetState(SCE_B_DEFAULT);}
          if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT);}
          break; }


        case SCE_B_COMMENT: {
          if (sc.atLineEnd) {sc.SetState(SCE_B_DEFAULT);}      // reset to default as line end
          break; }


      }                                        // end switch (sc.state)


//**********************************************************************************************
// we are in default state, determine if a new state should be entered:
//**********************************************************************************************
        if (sc.state == SCE_B_DEFAULT) {
          if ((sc.ch == '/') && (sc.chNext == '/')) {sc.SetState(SCE_B_COMMENT);}
          else if (sc.ch == '\"') {sc.SetState(SCE_B_STRING);}
          else if (sc.ch == '0' && tolower(sc.chNext) == 'x') {sc.SetState(SCE_B_NUMBER);}
          else if (IsADigit(sc.ch) || (sc.ch == '.' && IsADigit(sc.chNext))) {sc.SetState(SCE_B_NUMBER);}
          else if (IsAWordStart(sc.ch)) {sc.SetState(SCE_B_KEYWORD);} // make every new word a keyword first
          else if (sc.ch == '#') {sc.SetState(SCE_B_KEYWORD);}
          else if (isoperator(static_cast<char>(sc.ch)) || (sc.ch == '\\')) {sc.SetState(SCE_B_OPERATOR);}
        }

    }                                            // next
    sc.Complete();
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// rules:

// if level <= SC_FOLDLEVELBASE then level for these is SC_FOLDLEVELBASE - 1
// if level > SC_FOLDLEVELBASE then level for these is level + 1
// this way you can enclose SC_FOLDLEVELBASE lines (sub/function .., type enum
//   '->    //+*....   increase fold level (set a userdefined folding point, must be left aligned, no leading spaces)
//   '->    //-*....   decrease fold level

//ID accelerators (begin/end)  level = SC_FOLDLEVELBASE
//
//ID dialog (begin/end) 
//ID dialogex (begin/end) 
//
//ID menu (begin/end) 
//ID menuex (begin/end) 
//
//ID messagetable ??? -> no
//
// begin/end/{/} -> level += 1
//
//ID rcdata (begin/end) 
//
// stringtable (begin/end)
//
//ID versioninfo (begin/end) 


//**********************************************************************************************
// fold level is number between 0 and 0xFFF ("& SC_FOLDLEVELNUMBERMASK" isolates this level from style byte)
// | SC_FOLDLEVELHEADERFLAG adds a fold switch to this line
//**********************************************************************************************






static void FoldRCDoc(unsigned int startPos, int length, int, WordList *[], Accessor &styler) {
//**********************************************************************************************
// handle folding
//**********************************************************************************************
if( styler.GetPropertyInt("fold") == 0 ) return; // No folding enabled, no reason to continue...


char chU;                                        // current char (uppercase)
char chnU;                                       // next char (uppercase)
unsigned int header = 0;                         // set header flag ?
unsigned int first  = startPos;                  // is it the first pos in line ?

unsigned int endPos = startPos + length;         // scan until here
int lineCurrent     = styler.GetLine(startPos);  // current line # 
int lc              = SC_FOLDLEVELBASE;          // init to default


    if (lineCurrent > 0) {
		lc = styler.LevelAt(lineCurrent) & SC_FOLDLEVELNUMBERMASK; }       //get current level, mask out flags
    int ln = lc;


char ch = ' ';
char chNext   = styler[startPos];                // first char
int  fNewLine = 0;                               // flags



//char m[100] = "initial:                     " ;
//char m1[100];
//_itoa(lc, &m[10] , 10);                          //current level
//_itoa(lineCurrent, m1 , 10);                     //current line
//strcat(m, "  -  ");
//strcat(m, m1);
//OutputDebugStringA(m);                           //test


//********************************************************************************
// main for/next loop
//********************************************************************************


    for (unsigned int i = startPos; i < endPos; i++) {               // loop through chars
      ch = chNext;                          // current char
      chNext = styler.SafeGetCharAt(i + 1);      // next char
      chU  = ch & 95;                            // current char (uppercase)
      chnU = chNext & 95;                        // next char (uppercase)


      if (fNewLine == 0) {                            // Begin of a new line, the first non blank is the first token
//OutputDebugStringA("newline");
        if (ch == ' ')  {goto l1; }              // skip spaces
        if (ch == '\t') {goto l1; }              // skip tabs

        fNewLine = 1;                        // now we are at the first token

        if (ch == '/')  {
          if (i == first) {                      // only, if first in line (only leftaligned, ignore second and further pos)
            if (chNext == '/') {
              if (styler.SafeGetCharAt(i + 3) == '*') {
                if (styler.SafeGetCharAt(i + 2) == '+') {   // "'+*..." (user fold start)
                  if (lc > SC_FOLDLEVELBASE) {              //(regular) user fold point
                    ln = lc + 1; }
                  else {                           //at SC_FOLDLEVELBASE, step back onr level (enclose SC_FOLDLEVELBASE points)
                    lc = SC_FOLDLEVELBASE - 1;
                    ln = lc + 1; }

			      header = SC_FOLDLEVELHEADERFLAG;
                  fNewLine = 3;
			      goto l1; }

                if (styler.SafeGetCharAt(i + 2) == '-') {   // "'-*..." (user fold end)
                  fNewLine = 3;
                  ln = lc - 1;
                  goto l1; }}}}}


        else if (ch == '{') {                    // "{"
          if ((chNext == ' ') || (chNext == '\r')) {
            header = SC_FOLDLEVELHEADERFLAG;
            fNewLine = 3;
            ln = lc + 1;
            goto l1; }}

          
        else if (ch == '}') {                    // "}"
          if ((chNext == ' ') || (chNext == '\r')) {
            fNewLine = 3;
            ln = lc - 1;
            goto l1; }}


//**********************************************************************************************
// check first nonspace char (A-Z uppercase)
//**********************************************************************************************
        switch (chU) {                      
          case 'E': {
            if (chnU == 'N') {
               if (MatchUpperCase(styler,i,"END")) {
                 if ((styler.SafeGetCharAt(i+3) == ' ') || (styler.SafeGetCharAt(i+3) == '\r')) {
                   fNewLine = 3;

//_itoa(lc, m , 10);
//strcat(m, " - end");
//OutputDebugStringA(m);



                   ln = lc - 1;
                   goto l1; }}}

            break; }

              
          case 'B': {
            if (chnU == 'E') {
               if (MatchUpperCase(styler,i,"BEGIN")) {
                 if ((styler.SafeGetCharAt(i+5) == ' ') || (styler.SafeGetCharAt(i+5) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;

//_itoa(lc, m , 10);
//strcat(m, " - begin");
//OutputDebugStringA(m);




                   fNewLine = 3;
                   ln = lc + 1;
                   goto l1; }}}

            break; }


          case 'S': {
            if (chnU == 'T') {
               if (MatchUpperCase(styler,i,"STRINGTABLE")) {
                 if ((styler.SafeGetCharAt(i+11) == ' ') || (styler.SafeGetCharAt(i+11) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   fNewLine = 3;
                   ln = lc + 1;


//_itoa(ln, m , 10);
//strcat(m, " - after stringt");
//OutputDebugStringA(m);




                   goto l1; }}}

            break; }
        }                       // end switch
      }


      else if (fNewLine == 1) {  
        if ((ch == ' ') || (ch == '\t')) {goto l1;}
        else {
          fNewLine = 2;
          goto l2; }}
        

                         
      else if (fNewLine == 2) {  
        if ((ch == ' ') || (ch == '\t')) {goto l1;}
        if ((ch == '\r') || (ch == '\n')) {goto l2;}
        
//**********************************************************************************************
// check second token (A-Z uppercase)
//**********************************************************************************************
//ID accelerators (begin/end)  level = SC_FOLDLEVELBASE
//ID dialog (begin/end) 
//ID dialogex (begin/end) 
//ID menu (begin/end) 
//ID menuex (begin/end) 
//ID rcdata (begin/end) 
//ID versioninfo (begin/end) 
//**********************************************************************************************
        switch (chU) {                      // check for keywords
          case 'A': {
            if (chnU == 'C') {
               if (MatchUpperCase(styler,i-1," ACCELERATORS")) {
                 if ((styler.SafeGetCharAt(i+12) == ' ') || (styler.SafeGetCharAt(i+12) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   ln = lc + 1;
                   fNewLine = 3;
                   goto l1; }}}

            break; }
          
        
          case 'D': {
            if (chnU == 'I') {
               if (MatchUpperCase(styler,i-1," DIALOGEX")) {
                 if ((styler.SafeGetCharAt(i+8) == ' ') || (styler.SafeGetCharAt(i+8) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   fNewLine = 3;
                   ln = lc + 1;
                   goto l1; }}
            
               else if (MatchUpperCase(styler,i-1," DIALOG")) {
                 if ((styler.SafeGetCharAt(i+6) == ' ') || (styler.SafeGetCharAt(i+6) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   fNewLine = 3;

//OutputDebugStringA("Dialog");
                   ln = lc + 1;
                   goto l1; }}}
            
            break; }
          
        
          case 'M': {
            if (chnU == 'E') {
               if (MatchUpperCase(styler,i-1," MENUEX")) {
                 if ((styler.SafeGetCharAt(i+6) == ' ') || (styler.SafeGetCharAt(i+6) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   fNewLine = 3;
                   ln = lc + 1;
                   goto l1; }}
            
               else if (MatchUpperCase(styler,i-1," MENU")) {
                 if ((styler.SafeGetCharAt(i+4) == ' ') || (styler.SafeGetCharAt(i+4) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   fNewLine = 3;
                   ln = lc + 1;
                   goto l1; }}}
            
            break; }
          
        
          case 'R': {
            if (chnU == 'C') {
               if (MatchUpperCase(styler,i-1," RCDATA")) {
                 if ((styler.SafeGetCharAt(i+6) == ' ') || (styler.SafeGetCharAt(i+6) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   ln = lc + 1;
                   fNewLine = 3;
                   goto l1; }}}

            break; }
          
        
          case 'V': {
            if (chnU == 'E') {
               if (MatchUpperCase(styler,i-1," VERSIONINFO")) {
                 if ((styler.SafeGetCharAt(i+11) == ' ') || (styler.SafeGetCharAt(i+11) == '\r')) {
                   header = SC_FOLDLEVELHEADERFLAG;                   //fold point
                   lc = SC_FOLDLEVELBASE;                             //always at base level
                   ln = lc + 1;
                   fNewLine = 3;
                   goto l1; }}}

            break; }

          }                                        // end switch (ch)
        }                                          // end if (fNewLine = 2)


//**********************************************************************************************
// check for line end
//**********************************************************************************************


        //if (ch == '\r') {                        //carriage return -> line end reached
        //  if (chNext == '\n') {i = i + 1; }      //skip line feed
        //    goto l2; }                           //handle line end


l2: ;
        if (ch == '\n') {                        //line feed -> line end reached
          styler.SetLevel(lineCurrent, ((lc & SC_FOLDLEVELNUMBERMASK) | header));                //set level for current line


//_itoa(lineCurrent, m , 10);
//strcat(m, " - set");
//OutputDebugStringA(m);



          lineCurrent++;                         //increase line #
          lc = ln;                               //current level = next level
          fNewLine = 0;                          //a new line starts
          first = i + 1;                         //first = pos of first char in this new line
          header = 0; }


l1: ;
    }                                            // next


    if (ch != '\n') {                            //last char might not be a line feed, (no level set for this line)
      styler.SetLevel(lineCurrent, ((lc & SC_FOLDLEVELNUMBERMASK) | header)); } //set level for current line


    lineCurrent++;                               //sytle next line (level could be different)
    styler.SetLevel(lineCurrent, (ln & SC_FOLDLEVELNUMBERMASK));     //set level for next line

}


//**********************************************************************************************


static const char * const rcWordListDesc[] = {
    "Keywords",
    0
};

LexerModule lmRC(SCLEX_RC, ColouriseRCDoc, "RC", FoldRCDoc, rcWordListDesc);
